//
// Created by philipp on 30.11.17.
//

#ifndef PROTOCOL_FOLLOWERVEHICLE_H
#define PROTOCOL_FOLLOWERVEHICLE_H


#include <mutex>
#include "NetworkingLib/DatagramReceiver.h"
#include "NetworkingLib/DatagramSender.h"
#include "Vehicle.h"
#include "PlatoonMessage.h"
#include "Utils.h"

namespace platoonProtocol
{

class FollowerVehicle : public Vehicle
{
private:
    struct PrivateTag
    {
    };

public:
    static constexpr networking::time::Duration RESPONSE_TIMEOUT = std::chrono::milliseconds(3000);
    static constexpr networking::time::Duration HEARTBEAT_INTERVAL = std::chrono::milliseconds(200);
    static constexpr networking::time::Duration BROADCAST_TIMEOUT = std::chrono::milliseconds(500);

    using Ptr = std::shared_ptr<FollowerVehicle>;

    static Ptr create(networking::Networking & net, const NetworkInfo & info);

    // This should not be used outside.
    FollowerVehicle(PrivateTag, networking::Networking & net, const NetworkInfo & info);

    void stop() override;

    PlatoonConfig getPlatoonConfig() const noexcept;

    void setOnPlatoonConfigUpdatedCallback(const Callback & callback)
    { onPlatoonConfigUpdatedCallback = callback; }

protected:
    void doCreatePlatoon() override;

    void doRunPlatoon() override;

    void doLeavePlatoon() override;

    void receiveMessage(const PlatoonMessage & message) override;

    Ptr shared_from_this();

private:
    VehicleId leader;
    networking::time::Timer::Ptr responseTimer;
    networking::time::Timer::Ptr broadcastTimer;
    networking::time::Timer::Ptr heartbeatTimer;
    Callback onPlatoonConfigUpdatedCallback;

    void sendPlatoonCreateRequest();

    void receivePlatoonCreateRequest(const PlatoonMessage & request);

    void receiveResponse(const PlatoonMessage & response);

    void startHeartbeat();

    void stopHeartbeat();

    void startBroadcastTimer();

    void stopBroadcastTimer();

    void receiveBroadcast(const PlatoonMessage & broadcast);

    void sendLeavePlatoonMessage();
};

}

#endif //PROTOCOL_FOLLOWERVEHICLE_H
